/*
 * model1.c
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995-97)
 * this file belongs to the PUZZLE version 3.0 distribution
 *
 */


/* definitions */
#define EXTERN extern

/* prototypes */
#include <stdio.h>
#include "util.h"
#include "ml.h"

/* number of states of the selected model */ 
int gettpmradix()
{
	if (data_optn == 0) { /* nucleotides */
		if (nuc_optn) return 4;
		if (SH_optn) return 16;
	} else { /* amino acids */
		return 20;
	}
	return 1;
}

/* relative transition frequencies */
void rtfdata(dmatrix a, double *f)
{
	double alp, beta, alpy, alpr;
	dmatrix b;
	int i, j;

	if (data_optn == 0) { /* nucleotides */

		if (nuc_optn) { /* 4x4 nucleotides */
			beta = 1.0;
			alp = 2.0*TSparam;
			alpr = (alp * 2.0) / (YRparam + 1.0);
			alpy = YRparam * alpr;
			a[0][0] =  0.0; a[0][1] = beta; a[0][2] = alpr; a[0][3] = beta;
			a[1][0] = beta; a[1][1] =  0.0; a[1][2] = beta; a[1][3] = alpy;
			a[2][0] = alpr; a[2][1] = beta; a[2][2] =  0.0; a[2][3] = beta;
			a[3][0] = beta; a[3][1] = alpy; a[3][2] = beta; a[3][3] =  0.0;
			f[0] = 0.25; f[1] = 0.25; f[2] = 0.25; f[3] = 0.25;
		}
		
		if (SH_optn) { /* 16x16 nucleotides */
		
			alp = 2.0*TSparam;
	
			a[0][0] = 0; a[0][1] = 1; a[0][2] = alp; a[0][3] = 1;
			a[0][4] = 1; a[0][5] = 0; a[0][6] = 0; a[0][7] = 0;
			a[0][8] = alp; a[0][9] = 0; a[0][10] = 0; a[0][11] = 0;
			a[0][12] = 1; a[0][13] = 0; a[0][14] = 0; a[0][15] = 0;
			a[1][0] = 1; a[1][1] = 0; a[1][2] = 1; a[1][3] = alp;
			a[1][4] = 0; a[1][5] = 1; a[1][6] = 0; a[1][7] = 0;
			a[1][8] = 0; a[1][9] = alp; a[1][10] = 0; a[1][11] = 0;
			a[1][12] = 0; a[1][13] = 1; a[1][14] = 0; a[1][15] = 0;
			a[2][0] = alp; a[2][1] = 1; a[2][2] = 0; a[2][3] = 1;
			a[2][4] = 0; a[2][5] = 0; a[2][6] = 1; a[2][7] = 0;
			a[2][8] = 0; a[2][9] = 0; a[2][10] = alp; a[2][11] = 0;
			a[2][12] = 0; a[2][13] = 0; a[2][14] = 1; a[2][15] = 0;
			a[3][0] = 1; a[3][1] = alp; a[3][2] = 1; a[3][3] = 0;
			a[3][4] = 0; a[3][5] = 0; a[3][6] = 0; a[3][7] = 1;
			a[3][8] = 0; a[3][9] = 0; a[3][10] = 0; a[3][11] = alp;
			a[3][12] = 0; a[3][13] = 0; a[3][14] = 0; a[3][15] = 1;
			a[4][0] = 1; a[4][1] = 0; a[4][2] = 0; a[4][3] = 0;
			a[4][4] = 0; a[4][5] = 1; a[4][6] = alp; a[4][7] = 1;
			a[4][8] = 1; a[4][9] = 0; a[4][10] = 0; a[4][11] = 0;
			a[4][12] = alp; a[4][13] = 0; a[4][14] = 0; a[4][15] = 0;
			a[5][0] = 0; a[5][1] = 1; a[5][2] = 0; a[5][3] = 0;
			a[5][4] = 1; a[5][5] = 0; a[5][6] = 1; a[5][7] = alp;
			a[5][8] = 0; a[5][9] = 1; a[5][10] = 0; a[5][11] = 0;
			a[5][12] = 0; a[5][13] = alp; a[5][14] = 0; a[5][15] = 0;
			a[6][0] = 0; a[6][1] = 0; a[6][2] = 1; a[6][3] = 0;
			a[6][4] = alp; a[6][5] = 1; a[6][6] = 0; a[6][7] = 1;
			a[6][8] = 0; a[6][9] = 0; a[6][10] = 1; a[6][11] = 0;
			a[6][12] = 0; a[6][13] = 0; a[6][14] = alp; a[6][15] = 0;
			a[7][0] = 0; a[7][1] = 0; a[7][2] = 0; a[7][3] = 1;
			a[7][4] = 1; a[7][5] = alp; a[7][6] = 1; a[7][7] = 0;
			a[7][8] = 0; a[7][9] = 0; a[7][10] = 0; a[7][11] = 1;
			a[7][12] = 0; a[7][13] = 0; a[7][14] = 0; a[7][15] = alp;
			a[8][0] = alp; a[8][1] = 0; a[8][2] = 0; a[8][3] = 0;
			a[8][4] = 1; a[8][5] = 0; a[8][6] = 0; a[8][7] = 0;
			a[8][8] = 0; a[8][9] = 1; a[8][10] = alp; a[8][11] = 1;
			a[8][12] = 1; a[8][13] = 0; a[8][14] = 0; a[8][15] = 0;
			a[9][0] = 0; a[9][1] = alp; a[9][2] = 0; a[9][3] = 0;
			a[9][4] = 0; a[9][5] = 1; a[9][6] = 0; a[9][7] = 0;
			a[9][8] = 1; a[9][9] = 0; a[9][10] = 1; a[9][11] = alp;
			a[9][12] = 0; a[9][13] = 1; a[9][14] = 0; a[9][15] = 0;
			a[10][0] = 0; a[10][1] = 0; a[10][2] = alp; a[10][3] = 0;
			a[10][4] = 0; a[10][5] = 0; a[10][6] = 1; a[10][7] = 0;
			a[10][8] = alp; a[10][9] = 1; a[10][10] = 0; a[10][11] = 1;
			a[10][12] = 0; a[10][13] = 0; a[10][14] = 1; a[10][15] = 0;
			a[11][0] = 0; a[11][1] = 0; a[11][2] = 0; a[11][3] = alp;
			a[11][4] = 0; a[11][5] = 0; a[11][6] = 0; a[11][7] = 1;
			a[11][8] = 1; a[11][9] = alp; a[11][10] = 1; a[11][11] = 0;
			a[11][12] = 0; a[11][13] = 0; a[11][14] = 0; a[11][15] = 1;
			a[12][0] = 1; a[12][1] = 0; a[12][2] = 0; a[12][3] = 0;
			a[12][4] = alp; a[12][5] = 0; a[12][6] = 0; a[12][7] = 0;
			a[12][8] = 1; a[12][9] = 0; a[12][10] = 0; a[12][11] = 0;
			a[12][12] = 0; a[12][13] = 1; a[12][14] = alp; a[12][15] = 1;
			a[13][0] = 0; a[13][1] = 1; a[13][2] = 0; a[13][3] = 0;
			a[13][4] = 0; a[13][5] = alp; a[13][6] = 0; a[13][7] = 0;
			a[13][8] = 0; a[13][9] = 1; a[13][10] = 0; a[13][11] = 0;
			a[13][12] = 1; a[13][13] = 0; a[13][14] = 1; a[13][15] = alp;
			a[14][0] = 0; a[14][1] = 0; a[14][2] = 1; a[14][3] = 0;
			a[14][4] = 0; a[14][5] = 0; a[14][6] = alp; a[14][7] = 0;
			a[14][8] = 0; a[14][9] = 0; a[14][10] = 1; a[14][11] = 0;
			a[14][12] = alp; a[14][13] = 1; a[14][14] = 0; a[14][15] = 1;
			a[15][0] = 0; a[15][1] = 0; a[15][2] = 0; a[15][3] = 1;
			a[15][4] = 0; a[15][5] = 0; a[15][6] = 0; a[15][7] = alp;
			a[15][8] = 0; a[15][9] = 0; a[15][10] = 0; a[15][11] = 1;
			a[15][12] = 1; a[15][13] = alp; a[15][14] = 1; a[15][15] = 0;
			
			for (i = 0; i < 16; i++) f[i] = 0.0625;
		}
	} else { /* amino acids */
		if (Dayhf_optn) { /* Dayhoff model */
			b = new_dmatrix(20,20);
			jttdata(b, f);
			dyhfdata(a, f);
			for (i = 0; i < 20; i++) {
				for (j = 0; j < 20; j++) {
					if (a[i][j] == 0.0) {
						a[i][j] = b[i][j] * (1572.0 / 59190.0);
						if (a[i][j] > 1.0) a[i][j] = 1.0;
					}
				}
				a[i][i] = 0.0;
			}
			free_dmatrix(b);
		} else if (Jtt_optn) { /* JTT model */
			jttdata(a, f);
		} else { /* mtREV model */
			mtrevdata(a, f);
		}
	}
}

/* transform letter codes to state numbers */
int code2int(cvector c)
{	if (data_optn == 0) { /* nucleotides */
		if (nuc_optn) { /* 4x4 */
			switch (c[0]) {
				case 'A': return 0;
				case 'C': return 1;
				case 'G': return 2;
				case 'T': return 3;
				case 'U': return 3;
				default : return 4;
			}
		}
		if (SH_optn) { /* 16x16 */
			if (c[0] == 'A') {
				switch (c[1]) {
					case 'A': return 0; /* AA */
					case 'C': return 1; /* AC */
					case 'G': return 2; /* AG */
					case 'T': return 3; /* AT */
					case 'U': return 3; /* AT */
					default:  return 16;
				}
			}
			if (c[0] == 'C') {
				switch (c[1]) {
					case 'A': return 4; /* CA */
					case 'C': return 5; /* CC */
					case 'G': return 6; /* CG */
					case 'T': return 7; /* CT */
					case 'U': return 7; /* CT */
					default:  return 16;
				}
			}
			if (c[0] == 'G') {
				switch (c[1]) {
					case 'A': return 8; /* GA */
					case 'C': return 9; /* GC */
					case 'G': return 10; /* GG */
					case 'T': return 11; /* GT */
					case 'U': return 11; /* GT */
					default:  return 16;
				}
			}
			if (c[0] == 'T' || c[0] == 'U') {
				switch (c[1]) {
					case 'A': return 12; /* TA */
					case 'C': return 13; /* TC */
					case 'G': return 14; /* TG */
					case 'T': return 15; /* TT */
					case 'U': return 15; /* TT */
					default:  return 16;
				}
			}
			return 16;
		}
	} else { /* amino acids */
		switch (c[0]) {
			case 'A': return 0;
			case 'C': return 4;
			case 'D': return 3;
			case 'E': return 6;
			case 'F': return 13;
			case 'G': return 7;
			case 'H': return 8;
			case 'I': return 9;
			case 'K': return 11;
			case 'L': return 10;
			case 'M': return 12;
			case 'N': return 2;
			case 'P': return 14;
			case 'Q': return 5;
			case 'R': return 1;
			case 'S': return 15;
			case 'T': return 16;
			case 'V': return 19;
			case 'W': return 17;
			case 'Y': return 18;
			default : return 20;
		}
	}
	return 0;
}

/* return letter code belonging to state number */
char *int2code(int s)
{
	if (data_optn == 0) { /* nucleotides */
		if (nuc_optn) { /* 4x4 */
			switch (s) {
				case 0: return "A";
				case 1: return "C";
				case 2: return "G";
				case 3: return "T";
				default : return "?";
			}
		}
		if (SH_optn) { /* 16x16 */
			switch (s) {
				case 0: return "AA";
				case 1: return "AC";
				case 2: return "AG";
				case 3: return "AT";
				case 4: return "CA";
				case 5: return "CC";
				case 6: return "CG";
				case 7: return "CT";
				case 8: return "GA";
				case 9: return "GC";
				case 10: return "GG";
				case 11: return "GT";
				case 12: return "TA";
				case 13: return "TC";
				case 14: return "TG";
				case 15: return "TT";
				default : return "??";
			}
		}
	} else { /* amino acids */
		switch (s) {
			case 0: return "A";
			case 1: return "R";
			case 2: return "N";
			case 3: return "D";
			case 4: return "C";
			case 5: return "Q";
			case 6: return "E";
			case 7: return "G";
			case 8: return "H";
			case 9: return "I";
			case 10: return "L";
			case 11: return "K";
			case 12: return "M";
			case 13: return "F";
			case 14: return "P";
			case 15: return "S";
			case 16: return "T";
			case 17: return "W";
			case 18: return "Y";
			case 19: return "V";
			default : return "?";
		}
	}
	return "?";
}
